use anyhow::Result;
use serialport::*;
use std::io::prelude::*;
use std::thread::sleep;
use std::time::Duration;

pub struct SerialService {
    port: Box<dyn SerialPort>,
}

impl SerialService {
    pub fn new(port_name: &str) -> Result<Box<Self>> {
        let port = serialport::new(port_name, 115_200)
            .timeout(Duration::from_millis(200))
            .open()?;
        Ok(Box::new(SerialService { port }))
    }

    pub fn get_ports() -> Result<Vec<String>> {
        let portinfos = serialport::available_ports()?;
        let mut portnames: Vec<String> = Vec::new();
        for portinfo in portinfos {
            portnames.push(portinfo.port_name);
        }
        Ok(portnames)
    }

    pub fn send_command<T: for<'a> AsRef<str>>(&mut self, command: T, timeout: Option<u64>) -> Result<String, String> {
        let command_str = command.as_ref();
        let command = format!("{}\r\n", command_str);
        let timeout = timeout.unwrap_or(200);
        let _ = self.port.write_all(command.as_bytes());
        sleep(Duration::from_millis(timeout));
        let mut buffer: Vec<u8> = vec![0; 256];
        let size = match self.port.read(&mut buffer) {
            Ok(size) => { size }
            Err(msg) => { return Err(msg.to_string()); }
        };
        let mut resp = String::from_utf8_lossy(&buffer[..size]).to_string();
        resp = resp.replace("\r\n", "").replace(&command_str, "");
        Ok(resp)
    }

    pub fn send_command_with_target<T: for<'a> AsRef<str>>(&mut self, command: T, timeout: Option<u64>, target: T) -> Result<bool, String> {
        let target_str = target.as_ref();
        _ = return match self.send_command(command, timeout) {
            Ok(resp) => { Ok(resp.contains(target_str)) },
            Err(err) => Err(err)
        };
    }
}

#[test]
fn get_ports_test() {
    let ports = SerialService::get_ports().unwrap();
    for port in ports {
        println!("{:?}", port);
    }
}

#[test]
fn open_port_test() {
    let mut serial_service = SerialService::new("COM5").unwrap();
    let resp = serial_service.send_command("AT", Some(200)).unwrap();
    println!("{:?}", resp);
}

/*
pub struct SerialService {
    port: Box<dyn SerialPort>,
}

impl SerialService {
    pub fn new(port_name: &str) -> Result<Box<Self>, Box<dyn Error>> {
        let port = serial::open(port_name)?;
        return Ok(Box::new(SerialService { port: Box::new(port) }));
    }
    pub fn set_time_out(&mut self, timeout: u64) -> serial::Result<()> {
        self.port.set_timeout(Duration::from_millis(timeout))
    }

    pub fn set_options(&mut self, baud_rate: serial::BaudRate) -> serial::Result<()> {
        self.port.reconfigure(&|settings| {
            settings.set_baud_rate(baud_rate)?;
            settings.set_stop_bits(serial::Stop1);
            settings.set_parity(serial::ParityNone);
            settings.set_char_size(serial::Bits8);
            settings.set_flow_control(serial::FlowNone);
            Ok(())
        })?;
        self.port.set_timeout(Duration::from_millis(10))
    }

    fn _send_command<T: for<'a> AsRef<str>>(&mut self, cmd: T, delay: Option<u64>) -> serial::Result<String> {
        let command = format!("{}\r\n", cmd.as_ref());
        let delay = delay.unwrap_or(50);
        let mut buffer: Vec<u8> = vec![0; 256];
        let _ = self.port.write(command.as_bytes())?;
        sleep(Duration::from_millis(delay));
        let _ = match self.port.read(&mut buffer) {
            Ok(_) => {}
            Err(msg) => { return Ok(msg.to_string()) }
        };
        let resp = match std::str::from_utf8(&buffer) {
            Ok(str) => { str.to_string() }
            Err(_) => { String::new() }
        };
        Ok(resp)
    }

    pub fn send_command<T: for<'a> AsRef<str>>(&mut self, cmd: T, delay: Option<u64>, need_ok: Option<bool>) -> Result<String, String> {
        let need_ok = need_ok.unwrap_or(true);
        let resp = match self._send_command(cmd, delay) {
            Ok(resp) => resp,
            Err(err) => return Err(err.to_string())
        };
        if need_ok {
            if !resp.contains("OK") {
                return Err(resp);
            }
        }
        Ok(resp)
    }
}
 */